__all__ = [
    "CONFIG_PATH_ENV_NAME",
    "CORS_ORIGINS_ENV_NAME",
    "ROOT_PATH_ENV_NAME",
]

CONFIG_PATH_ENV_NAME = "QUALIBRATE_COMPOSITE_CONFIG_FILE"
ROOT_PATH_ENV_NAME = "QUALIBRATE_COMPOSITE_ROOT_PATH"
CORS_ORIGINS_ENV_NAME = "QUALIBRATE_COMPOSITE_CORS_ORIGINS"



import os
from pathlib import Path

import click
from qualibrate_config.vars import DEFAULT_CONFIG_FILENAME, QUALIBRATE_PATH

from qualibrate_composite.config import CONFIG_PATH_ENV_NAME
from qualibrate_composite.config.vars import (
    CORS_ORIGINS_ENV_NAME,
    ROOT_PATH_ENV_NAME,
)

try:
    from qualibrate_app.config import (
        CONFIG_PATH_ENV_NAME as QAPP_CONFIG_PATH_ENV_NAME,
    )
except ImportError:
    QAPP_CONFIG_PATH_ENV_NAME = None  # type: ignore[assignment]
try:
    from qualibrate_runner.config import (
        CONFIG_PATH_ENV_NAME as RUNNER_CONFIG_PATH_ENV_NAME,
    )
except ImportError:
    RUNNER_CONFIG_PATH_ENV_NAME = None  # type: ignore[assignment]


@click.command(name="start")
@click.option(
    "--config-path",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        path_type=Path,
    ),
    default=QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME,
    help="Path to `config.toml` file",
    show_default=True,
)
@click.option(
    "--reload", is_flag=True, hidden=True
)  # env QUALIBRATE_START_RELOAD
@click.option(
    "--port",
    type=int,
    default=8001,
    show_default=True,
    help="Application will be started on the given port",
)  # env QUALIBRATE_START_PORT
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    show_default=True,
    help="Application will be started on the given host",
)  # env QUALIBRATE_START_HOST
@click.option(
    "--cors-origin",
    type=str,
    multiple=True,
    help="CORS origin to use. Can be passed multiple times.",
)
@click.option(
    "--root-path",
    type=str,
    default="",
    help=(
        "Optional root path for the application if run as a sub-application."
    ),
)
def start_command(
    config_path: Path,
    port: int,
    host: str,
    reload: bool,
    cors_origin: list[str],
    root_path: str,
) -> None:
    config_path_str = str(config_path)
    os.environ[CONFIG_PATH_ENV_NAME] = config_path_str
    os.environ[ROOT_PATH_ENV_NAME] = root_path
    if len(cors_origin) != 0:
        os.environ[CORS_ORIGINS_ENV_NAME] = ",".join(cors_origin)
    if QAPP_CONFIG_PATH_ENV_NAME is not None:
        os.environ[QAPP_CONFIG_PATH_ENV_NAME] = str(config_path)
    if RUNNER_CONFIG_PATH_ENV_NAME is not None:
        os.environ[RUNNER_CONFIG_PATH_ENV_NAME] = str(config_path)

    from qualibrate_composite.app import main as app_main

    app_main(port=port, host=host, reload=reload, root_path=root_path)



import click
from qualibrate_config.cli.config import config_command

from qualibrate_composite.cli import start_command


@click.group()
def cli() -> None:
    pass


cli.add_command(config_command)
cli.add_command(start_command)


def main() -> None:
    cli(auto_envvar_prefix="QUALIBRATE")


if __name__ == "__main__":
    main()



import logging
from collections.abc import AsyncIterator
from contextlib import AsyncExitStack, asynccontextmanager
from importlib import metadata
from importlib.util import find_spec
from typing import Optional

from fastapi import FastAPI
from fastapi.routing import Mount
from packaging.requirements import Requirement
from packaging.version import Version

from qualibrate_composite.api.auth_middleware import (
    QualibrateAppAuthMiddleware,
    RunnerAuthMiddleware,
)
from qualibrate_composite.utils.logging_filter import EndpointFilter

# Frequently polled endpoints that should be filtered from logs
EXECUTION_STATUS_ENDPOINT = "/execution/is_running"
WORKFLOW_STATUS_ENDPOINT = "/execution/last_run/workflow/status"
WORKFLOW_HISTORY_ENDPOINT = "/execution/last_run/workflow/execution_history"
EXECUTION_LAST_RUN_STATUS_ENDPOINT = "/execution/last_run/status"
SNAPSHOTS_HISTORY_ENDPOINT = "/api/branch/main/snapshots_history"


def spawn_qualibrate_runner(app: FastAPI) -> None:
    try:
        from qualibrate_runner.app import app as runner_app
    except ImportError as ex:
        raise ImportError(
            "Can't import qualibrate_runner instance. "
            "Check that you have installed it."
        ) from ex
    logging.getLogger("uvicorn.access").addFilter(
        EndpointFilter(
            excluded_endpoints_starts=(
                EXECUTION_STATUS_ENDPOINT,
                WORKFLOW_STATUS_ENDPOINT,
                WORKFLOW_HISTORY_ENDPOINT,
                EXECUTION_LAST_RUN_STATUS_ENDPOINT,
                SNAPSHOTS_HISTORY_ENDPOINT,
            )
        )
    )
    runner_app.add_middleware(RunnerAuthMiddleware)
    app.mount("/execution", runner_app, name="qualibrate_runner")


def spawn_qualibrate_app(app: FastAPI) -> None:
    try:
        from qualibrate_app.app import app as qualibrate_app_app
    except ImportError as ex:
        raise ImportError(
            "Can't import qualibrate_app instance. "
            "Check that you have installed it."
        ) from ex

    qualibrate_app_app.add_middleware(QualibrateAppAuthMiddleware)
    app.mount("/", qualibrate_app_app, name="qualibrate_app")


def validate_runner_version_for_app() -> None:
    existing_version = Version(metadata.version("qualibrate-runner"))
    requirements_str = metadata.requires("qualibrate")
    if requirements_str is None:
        raise RuntimeError("There are no defined qualibrate dependencies.")
    requirements = map(Requirement, requirements_str)
    filtered_runner = filter(
        lambda r: r.name == "qualibrate-runner", requirements
    )
    requirement: Optional[Requirement] = next(filtered_runner, None)
    if requirement is None:
        logging.warning(
            "QUAlibrate-runner version is not recognized. This may be because "
            "the package was installed in editable mode. Cannot verify that "
            "version match."
        )
        return
    requirement_version_lst = list(iter(requirement.specifier))
    if (
        len(requirement_version_lst) != 1
        or not requirement_version_lst[0].operator == "=="
    ):
        raise RuntimeError(
            "Invalid required qualibrate-runner version format. "
            f"Your: {requirement.specifier}. "
            "Expected '==X.Y.Z'."
        )
    dep_version = Version(requirement_version_lst[0].version)
    if (
        existing_version.major == dep_version.major
        and existing_version.minor == dep_version.minor
        and existing_version.micro >= dep_version.micro
    ):
        return
    max_version = Version(f"{dep_version.major}.{dep_version.minor+1}.0")
    raise RuntimeError(
        f"Invalid qualibrate-runner version. Expected: '=={dep_version}'. "
        f"Allowed: '>={dep_version}, <{max_version}'. "
        f"Installed: {existing_version}. Please run "
        "'pip install --upgrade qualibrate' to ensure version compatibility."
    )


def spawn_qua_dashboards(app: FastAPI) -> None:
    if find_spec("qua_dashboards") is None:
        logging.warning(
            "qua_dashboards is not installed so the dashboards server "
            "is not started"
        )
        return
    path_prefix = app.root_path + "/dashboards"
    logging.getLogger("uvicorn.access").addFilter(
        EndpointFilter(excluded_endpoints_starts=(path_prefix,))
    )
    try:
        from qua_dashboards.app import create_app as qua_dashboard_create_app

        qua_dashboard_app = qua_dashboard_create_app(
            f"{path_prefix.rstrip('/')}/"
        )
    except Exception as ex:
        logging.exception("Can't import qua_dashboards", exc_info=ex)
        return
    from a2wsgi import WSGIMiddleware

    app.mount(
        "/dashboards",
        WSGIMiddleware(qua_dashboard_app.server),  # type: ignore[arg-type]
        name="qua_dashboards",
    )


@asynccontextmanager
async def app_lifespan(app: FastAPI) -> AsyncIterator[None]:
    async with AsyncExitStack() as stack:
        for route in filter(
            lambda r: isinstance(r, Mount) and isinstance(r.app, FastAPI),
            app.routes,
        ):
            await stack.enter_async_context(
                route.app.router.lifespan_context(app)  # type: ignore[attr-defined]
            )
        yield



from typing import Annotated, Optional

from fastapi import APIRouter, Body, Depends, Response, status
from fastapi.responses import JSONResponse
from qualibrate_config.models import QualibrateConfig

from qualibrate_composite.api.auth_middleware import encoded_password
from qualibrate_composite.config import get_settings

base_router = APIRouter()

AUTH_COOKIE_LIFETIME = 7 * 24 * 60 * 60


@base_router.get("/login_required")
def login_required(
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> bool:
    return settings.password is not None


@base_router.post("/login")
def login(
    password: Annotated[Optional[str], Body()] = None,
    *,
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> Response:
    if settings.password is None or password == settings.password:
        response = Response()
        password = encoded_password(password) if password else ""
        response.set_cookie(
            "Qualibrate-Token",
            password,
            max_age=AUTH_COOKIE_LIFETIME,
        )
    else:
        response = JSONResponse(
            {"detail": "Incorrect password"},
            status_code=status.HTTP_401_UNAUTHORIZED,
        )
        response.delete_cookie("Qualibrate-Token")
    return response
